(******************************************************************************
 * Copyright (c) 2012, Toyama&Aoto Laboratory, Tohoku University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 *  1. Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright 
 *     notice, this list of conditions and the following disclaimer in the 
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(******************************************************************************
 * file: rwtools/rwchecker/cr_diagram.sml
 * description: confluence check by decreasing diagram techniques using SMT Solver
 * author: AOTO Takahito
 * 
 * $Id: cr_diagram.sml,v 1.11 2012/05/14 04:48:13 aoto Exp $
 ******************************************************************************)


signature CR_DIAGRAM  = 
   sig
    val runDebug: bool ref
    val hasDecreasingCps: 
	string  (* minisat *)
	-> string  (* tmp dir *)
	-> (Term.term * Term.term) list 
	-> bool

    val fSymsOnVariablePathsInTerm: Term.term -> (Var.ord_key * (Fun.ord_key list)) list
end;

structure CrDiagram : CR_DIAGRAM= 
   struct

   local 
       open Term
       open Arith
(*        open Trs *)
(*        open Rewrite *)
(*        open Subst *)
(*        open Cr *)
       structure VS = VarSet
       structure VM = VarMap
       structure IM = IntMap
       structure FS = FunSet
       structure FM = FunMap
       structure FIM = FunIntMap
       structure IPM = IntPairMap
       structure SS = SortSet
       structure FIS = FunIntSet
       structure L = List
       structure LP = ListPair
       structure LU = ListUtil
       structure TS = TermSet
       open PrintUtil
   in

   val runDebug = ref false : bool ref
   fun debug f = if !runDebug then f () else ()
   exception CrDiagramError

   val decreasingDiagramVersion = ref 5
   (*********************************
       1: basic version
       2: counting designated function symbol's occurrence
       3: with context weight
       4: 3 + extended comparison
       5: 4 + extended context weight
   *********************************)



   (* $B4m81BP$N9gN.Ns(B($BMQ$$$k=q$-49$(5,B'(B)$B$N@8@.(B *)
(*    fun joinSequenceOfCps rs = *)
(*        let  *)
(* 	   val icps = L.@ (Cr.insideCriticalPairsWithIndexAndFs rs, *)
(* 			   Cr.outsideCriticalPairsInOnesideWithIndexAndFs rs) *)

(* 	   val maxCount = 5 *)

(* 	   exception NonJoinableCPFound *)

(* 	   val anslist =  *)
(* 	       List.map (fn ((m,n),(fs,c1,c2)) => *)
(* 			    let  *)
(* 				val _ = print "Critical Pair <" *)
(* 				val _ = print ((Term.toString c1) ^ ", " ^ (Term.toString c2) ^ ">") *)
(* 				val _ = print (" by Rules <" ^ (Int.toString m) *)
(* 					       ^ ", " ^ (Int.toString n) ^ ">") *)
(* 				val _ = print (" preceded by [" ^ *)
(* 					       (PrintUtil.prSeq Fun.toString fs) ^ "]\n") *)
(* 			    in *)
(* 				case Cr.findCommonReductWithIndexAndFs rs maxCount *)
(* 								       ((fs,m,c1),([],n,c2)) of *)
(* 				    NONE => raise NonJoinableCPFound *)
(* 				  | SOME (ms,ns) =>  *)
(* 				    let val (_::ms') = L.map (fn (fs,i,_) => (fs,i)) (rev ms) *)
(* 					val (_::ns') = L.map (fn (fs,i,_) => (fs,i)) (rev ns) *)
(* 					val _ = print (" joinable by a reduction of rules <["  *)
(* 						       ^ (PrintUtil.prSeq  *)
(* 							      (fn (fs,i) =>  *)
(* 								  "(["  *)
(* 								  ^ (PrintUtil.prSeq Fun.toString fs)  *)
(* 								  ^ "]," ^ (Int.toString i) ^ ")") *)
(* 							      ms') *)
(* 						       ^ "], ["  *)
(* 						       ^ (PrintUtil.prSeq  *)
(* 							      (fn (fs,i) =>  *)
(* 								  "(["  *)
(* 								  ^ (PrintUtil.prSeq Fun.toString fs)  *)
(* 								  ^ "]," ^ (Int.toString i) ^ ")") *)
(* 							      ns') *)
(* 						       ^ "]>\n") *)
(* 				    in *)
(* 					((fs,m),([],n),ms',ns') *)
(* 				    end *)
(* 			    end) *)
(* 			icps *)

(*        in *)
(* 	   SOME anslist *)
(*        end *)
(*        handle NonJoinableCPFound => NONE *)


   (* $B4m81BP$N9gN.Ns(B($BMQ$$$k=q$-49$(5,B'(B)$B$N@8@.(B *)
   fun joinSequenceOfCps rs =
       let 
 	   val icps = L.@ (Cr.insideCriticalPairsWithIndexAndFs rs, 
 			   Cr.outsideCriticalPairsInOnesideWithIndexAndFs rs)

	   val maxCount = 5

	   exception NonJoinableCPFound

	   val anslist = 
	       List.map (fn ((m,n),(fs,c1,c2)) =>
			    let 
				val _ = print "Critical Pair <"
				val _ = print ((Term.toString c1) ^ ", " ^ (Term.toString c2) ^ ">")
				val _ = print (" by Rules <" ^ (Int.toString m)
					       ^ ", " ^ (Int.toString n) ^ ">")
				val _ = print (" preceded by [" ^
					       (PrintUtil.prSeq Fun.toString fs) ^ "]\n")

				val joinSeqs = Cr.findCommonReductWithIndexAndFs rs maxCount
										 ((fs,m,c1),([],n,c2))
				(* non-joinablity check *)
				val _ = if null joinSeqs
					then raise NonJoinableCPFound
					else ()
			    in
				L.map 
				    (fn (ms,ns) => 
					let val (_::ms') = L.map (fn (fs,i,_) => (fs,i)) (rev ms)
					    val (_::ns') = L.map (fn (fs,i,_) => (fs,i)) (rev ns)
					    val _ = print (" joinable by a reduction of rules <[" 
							   ^ (PrintUtil.prSeq 
								  (fn (fs,i) => 
								      "([" 
								      ^ (PrintUtil.prSeq Fun.toString fs) 
								      ^ "]," ^ (Int.toString i) ^ ")")
								  ms')
							   ^ "], [" 
							   ^ (PrintUtil.prSeq 
								  (fn (fs,i) => 
								      "([" 
								      ^ (PrintUtil.prSeq Fun.toString fs) 
								      ^ "]," ^ (Int.toString i) ^ ")")
								  ns')
							   ^ "]>\n")
					in
					    ((fs,m),([],n),ms',ns')
					end)
				    joinSeqs
			    end)
			icps
			
       in
	   SOME anslist
       end
       handle NonJoinableCPFound => NONE

   (* $B4m81BP$N9gN.Ns(B($BMQ$$$k=q$-49$(5,B'(B)$B$N@8@.(B *)
   fun joinSequenceOfCps2 rs =
       let 
	   val icps0 = L.@ (Cr.insideCriticalPairsWithIndexAndPosFs rs,
			    Cr.outsideCriticalPairsInOnesideWithIndexAndPosFs rs)

(* 	   val icps = L.@ (Cr.insideCriticalPairsWithIndexAndFs rs, *)
(* 			   Cr.outsideCriticalPairsInOnesideWithIndexAndFs rs) *)

(* 	   val _ = L.app   *)
(* 		       (fn ((m,n),(ps,fs,c1,c2)) => *)
(* 			   (print "Critical Pair <"; *)
(* 			    print ((Term.toString c1) ^ ", " ^ (Term.toString c2) ^ ">"); *)
(* 			    print (" by Rules <" ^ (Int.toString m) *)
(* 				   ^ ", " ^ (Int.toString n) ^ ">"); *)
(* 			    print (" preceded by [" ^ *)
(* 				   (PrintUtil.prSeq Fun.toString fs) ^ "]\n"); *)
(* 			    print (" position [" ^ *)
(* 				   (PrintUtil.prSeq Int.toString ps) ^ "]\n"))) *)
(* 		       icps0  *)

	   val icps = L.map (fn (mn,(ps,fs,y,z)) => (mn,(LP.zip (fs,ps),y,z))) icps0

(*	   val maxCount = 5 *)
	   val maxCount = 5

	   exception NonJoinableCPFound

	   fun prFunInt (f,i) = "(" ^ (Fun.toString f) ^ "," 
				^ (Int.toString i) ^ ")"

	   val anslist = 
	       List.map (fn ((m,n),(fps,c1,c2)) =>
			    let 
				val _ = print "Critical Pair <"
				val _ = print ((Term.toString c1) ^ ", " ^ (Term.toString c2) ^ ">")
				val _ = print (" by Rules <" ^ (Int.toString m)
					       ^ ", " ^ (Int.toString n) ^ ">")
				val _ = print (" preceded by [" ^
					       (PrintUtil.prSeq prFunInt fps) ^ "]\n")

				val joinSeqs = Cr.findCommonReductWithIndexAndPosFs rs maxCount
										 ((fps,m,c1),([],n,c2))
				(* non-joinablity check *)
				val _ = if null joinSeqs
					then raise NonJoinableCPFound
					else ()
			    in
				L.map 
				    (fn (ms,ns) => 
					let val (_::ms') = L.map (fn (fps,i,_) => (fps,i)) (rev ms)
					    val (_::ns') = L.map (fn (fps,i,_) => (fps,i)) (rev ns)
					    val _ = print (" joinable by a reduction of rules <[" 
							   ^ (PrintUtil.prSeq 
								  (fn (fps,i) => 
								      "([" 
								      ^ (PrintUtil.prSeq prFunInt fps) 
								      ^ "]," ^ (Int.toString i) ^ ")")
								  ms')
							   ^ "], [" 
							   ^ (PrintUtil.prSeq 
								  (fn (fps,i) => 
								      "([" 
								      ^ (PrintUtil.prSeq prFunInt fps) 
								      ^ "]," ^ (Int.toString i) ^ ")")
								  ns')
							   ^ "]>\n")
					in
					    ((fps,m),([],n),ms',ns')
					end)
				    joinSeqs
			    end)
			icps
			
       in
	   SOME anslist
       end
       handle NonJoinableCPFound => NONE

   fun mkFsMap counterRef fs = 
       L.foldl
           (fn (f,en) => (counterRef := 1 + !counterRef;
                          FM.insert (en,f,!counterRef)))
           FM.empty 
           fs

   fun mkFpsMap counterRef fps = 
       L.foldl
           (fn (fi,en) => (counterRef := 1 + !counterRef;
                           FIM.insert (en,fi,!counterRef)))
           FIM.empty 
           fps

   fun lookupFsMap (fsMap,f) = 
       case FM.find (fsMap,f) of
           SOME n => n
         | NONE => (print ("looking up fsMap fails: (" 
                           ^ (Fun.toString f) ^ ")\n");
                    raise CrDiagramError)


   fun lookupFpsMap (fpsMap,(f,i)) = 
       case FIM.find (fpsMap,(f,i)) of
           SOME n => n
         | NONE => (print ("looking up fpsMap fails: (" 
                           ^ (Fun.toString f) ^ ","
                           ^ (Int.toString i) ^ ")\n");
                    raise CrDiagramError)

   fun encodeGtOnFs fsMap (fs,gs) =
       let val polyFs = L.map (fn f => [Var (lookupFsMap (fsMap,f))]) fs
	   val polyGs = L.map (fn g => [Var (lookupFsMap (fsMap,g))]) gs
       in Atom (Gt (polyFs,polyGs))
       end

   fun encodeGeOnFs fsMap (fs,gs) =
       let val polyFs = L.map (fn f => [Var (lookupFsMap (fsMap,f))]) fs
	   val polyGs = L.map (fn g => [Var (lookupFsMap (fsMap,g))]) gs
       in Atom (Ge (polyFs,polyGs))
       end

   fun encodeEqOnFs fsMap (fs,gs) =
       let val polyFs = L.map (fn f => [Var (lookupFsMap (fsMap,f))]) fs
	   val polyGs = L.map (fn g => [Var (lookupFsMap (fsMap,g))]) gs
       in Atom (Eq (polyFs,polyGs))
       end

   fun encodeGtOnFps fpsMap (fps,gps) =
       let val polyFps = L.map (fn fp => [Var (lookupFpsMap (fpsMap,fp))]) fps
	   val polyGqs = L.map (fn gq => [Var (lookupFpsMap (fpsMap,gq))]) gps
       in Atom (Gt (polyFps,polyGqs))
       end

   fun encodeGeOnFps fpsMap (fps,gps) =
       let val polyFps = L.map (fn fp => [Var (lookupFpsMap (fpsMap,fp))]) fps
	   val polyGqs = L.map (fn gq => [Var (lookupFpsMap (fpsMap,gq))]) gps
       in Atom (Ge (polyFps,polyGqs))
       end

   fun encodeEqOnFps fpsMap (fps,gps) =
       let val polyFps = L.map (fn fp => [Var (lookupFpsMap (fpsMap,fp))]) fps
	   val polyGqs = L.map (fn gq => [Var (lookupFpsMap (fpsMap,gq))]) gps
       in Atom (Eq (polyFps,polyGqs))
       end

   fun encodeGtOnRs (i,j) = Atom (Gt ([[Var (i+1)]],[[Var (j+1)]]))
   fun encodeGeOnRs (i,j) = Atom (Ge ([[Var (i+1)]],[[Var (j+1)]]))
   fun encodeEqOnRs (i,j) = Atom (Eq ([[Var (i+1)]],[[Var (j+1)]]))

   fun encodeGt1 fsMap lenRs ((fs,m),(gs,x)) =
       if (!decreasingDiagramVersion >= 4)
       then
	   Disj [encodeGtOnRs (m,x),
 		 Conj [encodeEqOnRs (m,x),
 		       encodeGtOnFs fsMap (fs,gs)],
 		 Conj [encodeEqOnRs (m,x),
 		       encodeEqOnFs fsMap (fs,gs),
		       encodeGtOnRs (lenRs+m,lenRs+x)]]
       else 
	   Disj [encodeGtOnFs fsMap (fs,gs),
 		 Conj [encodeEqOnFs fsMap (fs,gs),
		       encodeGtOnRs (m,x)]]

   fun encodeEq1 fsMap lenRs ((fs,m),(gs,x)) =
       if (!decreasingDiagramVersion >= 4)
       then
	   Conj [encodeEqOnRs (m,x),
		 encodeEqOnFs fsMap (fs,gs),
		 encodeEqOnRs (lenRs+m,lenRs+x)]
       else
	   Conj [encodeEqOnFs fsMap (fs,gs),
		 encodeEqOnRs (m,x)]

   fun encodeEq2 fpsMap lenRs ((fps,m),(gqs,x)) =
       Conj [encodeEqOnRs (m,x),
	     encodeEqOnFps fpsMap (fps,gqs), 
	     encodeEqOnRs (lenRs+m,lenRs+x)]
		   
   fun encodeGt2 fpsMap lenRs ((fps,m),(gqs,x)) =
       Disj [encodeGtOnRs (m,x),
	     Conj [encodeEqOnRs (m,x),
		   encodeGtOnFps fpsMap (fps,gqs)],
	     Conj [encodeEqOnRs (m,x),
		   encodeEqOnFps fpsMap (fps,gqs),
		   encodeGtOnRs (lenRs+m,lenRs+x)]]

   (* $B4m81BP$N9gN.>r7o$N7W;;(B *)       
   fun makeDecreasingCondition0 rs =
       let
	   fun makeDecreasingCond0 (_,n) (_,y) = encodeEqOnRs (n,y)
	   fun makeDecreasingCond1 ((_,m),ms) =
	       Conj (L.map (fn (_,x) => encodeGtOnRs (m,x)) ms)
	   fun makeDecreasingCond2 ((_,m),(_,n),ms) =
	       Conj (L.map (fn (_,x) =>
			       Disj [encodeGtOnRs (m,x), 
				     encodeGtOnRs (n,x)])
			   ms)
	   fun makeCondSub (m,n,ms) =
	       let val len = length ms
		   val indexes = L.tabulate (len, fn i => i)
	       in
		   Disj ((makeDecreasingCond2 (m,n,ms))
			 ::
			 (L.map (fn i => let val xs = L.take (ms,i)
					     val (y::ys) = L.drop (ms,i)
					 in Conj [makeDecreasingCond1 (m,xs),
						  makeDecreasingCond0 n y,
						  makeDecreasingCond2 (m,n,ys) ]
					 end)
				(L.tabulate (length ms, fn i => i))))
	       end

	   fun makeCond (m,n,ms,ns) = Conj [makeCondSub (m,n,ms), makeCondSub (n,m,ns)]

	   fun makeCond2 xss = Disj (L.map makeCond xss)

       in case joinSequenceOfCps rs of
	      NONE => False
	    | SOME seqs => Conj (L.map makeCond2 seqs)
       end

   (* $B4m81BP$N9gN.>r7o$N7W;;(B *)
   fun makeDecreasingCondition1 fsMap lenRs rs =
       let
	   fun makeDecreasingCond1 (m,ms) =
	       Conj (L.map (fn x => encodeGt1 fsMap lenRs (m,x)) ms)
	   fun makeDecreasingCond2 (m,n,ms) =
	       Conj (L.map (fn x =>
			       Disj [encodeGt1 fsMap lenRs (m,x), 
				     encodeGt1 fsMap lenRs (n,x)])
			   ms)
	   fun makeCondSub (m,n,ms) =
	       let val len = length ms
		   val indexes = L.tabulate (len, fn i => i)
	       in
		   Disj ((makeDecreasingCond2 (m,n,ms))
			 ::
			 (L.map (fn i => let val xs = L.take (ms,i)
					     val (y::ys) = L.drop (ms,i)
					 in Conj [makeDecreasingCond1 (m,xs),
						  encodeEq1 fsMap lenRs (n,y),
						  makeDecreasingCond2 (m,n,ys) ]
					 end)
				(L.tabulate (length ms, fn i => i))))
	       end

	   fun makeCond (m,n,ms,ns) = Conj [makeCondSub (m,n,ms), makeCondSub (n,m,ns)]

	   fun makeCond2 xss = Disj (L.map makeCond xss)

       in case joinSequenceOfCps rs of
	      NONE => False
	    | SOME seqs => Conj (L.map makeCond2 seqs)
       end

   (* $B4m81BP$N9gN.>r7o$N7W;;(B *)
   fun makeDecreasingCondition2 fpsMap lenRs rs =
       let
	   fun makeDecreasingCond1 (m,ms) =
	       Conj (L.map (fn x => encodeGt2 fpsMap lenRs (m,x)) ms)
	   fun makeDecreasingCond2 (m,n,ms) =
	       Conj (L.map (fn x =>
			       Disj [encodeGt2 fpsMap lenRs (m,x), 
				     encodeGt2 fpsMap lenRs (n,x)])
			   ms)
	   fun makeCondSub (m,n,ms) =
	       let val len = length ms
		   val indexes = L.tabulate (len, fn i => i)
	       in
		   Disj ((makeDecreasingCond2 (m,n,ms))
			 ::
			 (L.map (fn i => let val xs = L.take (ms,i)
					     val (y::ys) = L.drop (ms,i)
					 in Conj [makeDecreasingCond1 (m,xs),
						  encodeEq2 fpsMap lenRs (n,y),
						  makeDecreasingCond2 (m,n,ys) ]
					 end)
				(L.tabulate (length ms, fn i => i))))
	       end

	   fun makeCond (m,n,ms,ns) = Conj [makeCondSub (m,n,ms), makeCondSub (n,m,ns)]

	   fun makeCond2 xss = Disj (L.map makeCond xss)

       in case joinSequenceOfCps2 rs of
	      NONE => False
	    | SOME seqs => Conj (L.map makeCond2 seqs)

       end


  (* root $B$+$i(B $BJQ?t=P8=$=$l$>$l$X;j$k%Q%9>e$N4X?t5-9fNs(B *)
   fun fSymsOnVariablePathsInTerm (Term.Var (y,_)) = [(y,[])]
     | fSymsOnVariablePathsInTerm (Term.Fun (f,ts,_)) =
       L.map (fn (x,fs) => (x,f::fs))
	     (LU.mapAppend fSymsOnVariablePathsInTerm ts)

   fun makeNonOverlappingCondition fsMap (l,r) =
       let val vfsInL = fSymsOnVariablePathsInTerm l
	   val vfsInR = fSymsOnVariablePathsInTerm r
	   val vs = VS.listItems (Term.varSetInTerm l)
       in
	   Conj (L.map
		     (fn x => let val lfs = hd (L.mapPartial (fn (y,fs) => if Var.equal (x,y)
									   then SOME fs
									   else NONE) 
							     vfsInL)
			      in  (* duplicating $B$JJQ?t(B $B$O(B > $B!$(Bnon-duplicating $B$J$i(B >= *)
				  case (L.mapPartial (fn (y,fs) => if Var.equal (x,y)
								   then SOME fs
								   else NONE) 
						     vfsInR) of
				      [] => True
				    | (rfs::[]) => encodeGeOnFs fsMap (lfs,rfs)
				    | rfss => Conj (L.map 
						       (fn rfs => 
							   encodeGtOnFs fsMap (lfs,rfs))
						       rfss)
			      end)
		     vs)
       end


  (* root $B$+$i(B $BJQ?t=P8=$=$l$>$l$X;j$k%Q%9>e$N4X?t5-9fNs(B *)
   fun fSymArgsOnVariablePathsInTerm (Term.Var (y,_)) = [(y,[])]
     | fSymArgsOnVariablePathsInTerm (Term.Fun (f,ts,_)) =
       LU.mapAppend (fn i => 
			L.map (fn (x,fps) => (x,(f,i+1)::fps))
			      (fSymArgsOnVariablePathsInTerm (L.nth (ts,i))))
		    (L.tabulate (length ts, fn x => x))

   fun makeNonOverlappingCondition2 fpsMap (l,r) =
       let val vfpsInL = fSymArgsOnVariablePathsInTerm l
	   val vfpsInR = fSymArgsOnVariablePathsInTerm r
	   val vs = VS.listItems (Term.varSetInTerm l)
       in
	   Conj (L.map
		     (fn x => let val lfps = hd (L.mapPartial (fn (y,fps) => if Var.equal (x,y)
									     then SOME fps
									     else NONE) 
							      vfpsInL)
			      in  (* duplicating $B$JJQ?t(B $B$O(B > $B!$(Bnon-duplicating $B$J$i(B >= *)
				  case (L.mapPartial (fn (y,fps) => if Var.equal (x,y)
								    then SOME fps
								    else NONE) 
						     vfpsInR) of
				      [] => True
				    | (rfps::[]) => encodeGeOnFps fpsMap (lfps,rfps)
				    | rpfss => Conj (L.map 
						       (fn rfps => 
							   encodeGtOnFps fpsMap (lfps,rfps))
						       rpfss)
			      end)
		     vs)
       end

  (* $B4pK\%k!<%k%i%Y%j%s%0(B *)
   fun hasDecreasingCps0 smtSolverPath tmpDir rs =
       if (Trs.areLinearRules rs)
       then
	   let 
	       val lenRs = List.length rs
	       val counter = ref lenRs  (* integer variable $B$N%+%&%s%?(B *)
	       val _ = print (Int.toString (!counter))
	       val _ = print "\n"
	       val prop = makeDecreasingCondition0 rs
	       val (result,assign) = Solver.smtSolver smtSolverPath tmpDir (prop,!counter)
	       fun lookupVal i = case L.find (fn (j,_) => i = j) assign of 
				  SOME (_,w) =>  w
				| NONE => (print ("no assingment for x" ^ (Int.toString i) ^ "\n");
					   raise CrDiagramError)
	       fun countOrder len = 
		   if result
		   then
		   let val sorted =
			   ListMergeSort.sort (fn ((i1,i2),(j1,j2)) => 
						  i1 > j1
						  orelse ((i1 = j1) andalso j1 > j2))
					      (L.tabulate (lenRs, fn i => (lookupVal (len+i+1),i+1)))
(* 		       val _ = print "\n" *)
(* 		       val _ = L.app  *)
(* 			       (fn (i,j) => (print ((Int.toString i) ^ ":=:" ^ (Int.toString j)))) *)
(* 			       sorted *)
(* 		       val _ = print "\n" *)
		   in case sorted of
			  [] => (0,"")
			| ((x1,x2)::xs) => L.foldl
						(fn ((i,j),(pre,str)) => 
						    if i > pre
						    then (i, (Int.toString j) ^  ">" ^ str)
						    else (i, (Int.toString j) ^  "," ^ str))
						(x1, Int.toString x2)
						xs
		   end
		   else (0,"")

	       val _ = if result then print ("Satisfiable by " ^ (#2 (countOrder 0)) ^ "\n")  else ()



	   in 
	       result
	   end
       else false

  (* $B=E$_IU$-%k!<%k%i%Y%j%s%0(B *)
   fun hasDecreasingCps1 smtSolverPath tmpDir rs =
       if (Trs.areLeftLinearRules rs)
       then
	   let 
	       val lenRs = List.length rs
	       val counter = ref (lenRs + lenRs) (* integer variable $B$N%+%&%s%?(B *)
	       val _ = print (Int.toString (!counter))
	       val _ = print "\n"
	       val fs = FS.listItems (Trs.funSetInRules rs)
               val fsMap = mkFsMap counter fs (* fsMap: $B4X?t5-9f(B |-> $BJQ?tHV9f(B *)
(* 	       val _ = List.app (fn f => (print (Fun.toString f);  *)
(* 					  print (Int.toString (lookupFsMap (fsMap,f))); *)
(* 					  print "\n"))  fs *)
(* 	       val _ = print "\n" *)
(* 	       val _ = print (Int.toString (!counter)) *)
(* 	       val _ = print "\n" *)


	       (* $B4X?t5-9f$N=E$_$O(B1$B0J2<(B *)
(*  	       val prop0 = Conj (L.tabulate (!counter - lenRs,  *)
(*  					  fn i => (Atom (Ge ([[Const 1]], [[Var (i+lenRs)]]))))) *)

	       val prop1 = Conj (L.map (makeNonOverlappingCondition fsMap) rs)
	       val prop2 = makeDecreasingCondition1 fsMap lenRs rs
(*	       val prop = Conj [prop0,prop1,prop2] *)
	       val prop = if (!decreasingDiagramVersion) = 2
			  then
			      let (* $B4X?t5-9f$N=E$_$NOB$O(B1$B0J2<(B *)
    				  val prop0 = Atom (Ge ([[Const 1]],  
  							L.tabulate (!counter - lenRs, 
								 fn i => [Var (i+lenRs)]))) 
			      in Conj [prop0,prop1,prop2]
			      end
			  else Conj [prop1,prop2] 


	       val (result,assign) = Solver.smtSolver smtSolverPath tmpDir (prop,!counter)

	       fun lookupVal i = case L.find (fn (j,_) => i = j) assign of 
				  SOME (_,w) =>  w
				| NONE => (print ("no assingment for x" ^ (Int.toString i) ^ "\n");
					   raise CrDiagramError)

	       fun countOrder len = 
		   if result
		   then
		   let val sorted =
			   ListMergeSort.sort (fn ((i1,i2),(j1,j2)) => 
						  i1 > j1
						  orelse ((i1 = j1) andalso j1 > j2))
					      (L.tabulate (lenRs, fn i => (lookupVal (len+i+1),i+1)))
(* 		       val _ = print "\n" *)
(* 		       val _ = L.app  *)
(* 			       (fn (i,j) => (print ((Int.toString i) ^ ":=:" ^ (Int.toString j)))) *)
(* 			       sorted *)
(* 		       val _ = print "\n" *)
		   in case sorted of
			  [] => (0,"")
			| ((x1,x2)::xs) => L.foldl
						(fn ((i,j),(pre,str)) => 
						    if i > pre
						    then (i, (Int.toString j) ^  ">" ^ str)
						    else (i, (Int.toString j) ^  "," ^ str))
						(x1, Int.toString x2)
						xs
		   end
		   else (0,"")

	       val _ = if result then print ("Satisfiable by " ^ (#2 (countOrder 0)) ^ "; ")  else ()

	       val fws
		 = if result
		   then L.foldl
			    (fn (f,str) => 
				(str ^ (Fun.toString f ^ "(" ^
					(Int.toString (lookupVal (lookupFsMap (fsMap,f))))
					^ ")")))
			    ""
			    fs
		   else ""

	       val _ = if result 
		       then if (!decreasingDiagramVersion) <  4
			    then print (fws ^ "\n") 
			    else print (fws ^ "; ") 
		       else ()

	       val _ = if result 
			  andalso (!decreasingDiagramVersion) =  4
		       then print ((#2 (countOrder lenRs)) ^ "\n") 
		       else ()

	   in 
	       result
	   end
       else false

  (* $B3HD%%k!<%k%i%Y%j%s%0(B *)
   fun hasDecreasingCps2 smtSolverPath tmpDir rs =
       if (Trs.areLeftLinearRules rs)
       then
	   let 
	       val lenRs = List.length rs
	       val counter = ref (lenRs + lenRs) (* integer variable $B$N%+%&%s%?(B *)
(* 	       val _ = print (Int.toString (!counter)) *)
(* 	       val _ = print "\n" *)
	       val fargs = FIS.listItems (Trs.funAritySetInRules rs)
	       val fps = LU.mapAppend 
			     (fn (f,arg) => L.tabulate (arg, fn i=>(f,i+1)))
			     fargs
               val fpsMap = mkFpsMap counter fps (* fpsMap: ($B4X?t5-9f(B,int) |-> $BJQ?tHV9f(B *)
(* 	       val _ = List.app (fn (f,i) =>  *)
(* 				    (print ("(" ^ (Fun.toString f) ^ "," *)
(* 					    ^ (Int.toString i) ^ ") |-> "); *)
(* 				     print (Int.toString (lookupFpsMap (fpsMap,(f,i)))); *)
(* 				     print "\n"))  fps *)
(* 	       val _ = print "\n" *)
(* 	       val _ = print (Int.toString (!counter)) *)
(* 	       val _ = print "\n" *)
	       val prop1 = Conj (L.map (makeNonOverlappingCondition2 fpsMap) rs)
	       val prop2 = makeDecreasingCondition2 fpsMap lenRs rs
	       val prop = Conj [prop1,prop2]
	       val (result,assign) = Solver.smtSolver smtSolverPath tmpDir (prop,!counter)

	       fun lookupVal i = case L.find (fn (j,_) => i = j) assign of 
				  SOME (_,w) =>  w
				| NONE => (print ("no assingment for x" ^ (Int.toString i) ^ "\n");
					   raise CrDiagramError)

	       fun countOrder len = 
		   if result
		   then
		   let val sorted =
			   ListMergeSort.sort (fn ((i1,i2),(j1,j2)) => 
						  i1 > j1
						  orelse ((i1 = j1) andalso j1 > j2))
					      (L.tabulate (lenRs, fn i => (lookupVal (len+i+1),i+1)))
(* 		       val _ = print "\n" *)
(* 		       val _ = L.app  *)
(* 			       (fn (i,j) => (print ((Int.toString i) ^ ":=:" ^ (Int.toString j)))) *)
(* 			       sorted *)
(* 		       val _ = print "\n" *)
		   in case sorted of
			  [] => (0,"")
			| ((x1,x2)::xs) => L.foldl
						(fn ((i,j),(pre,str)) => 
						    if i > pre
						    then (i, (Int.toString j) ^  ">" ^ str)
						    else (i, (Int.toString j) ^  "," ^ str))
						(x1, Int.toString x2)
						xs
		   end
		   else (0,"")

	       val _ = if result then print ("Satisfiable by " ^ (#2 (countOrder 0)) ^ "; ")  else ()

	       val fps = if result
			 then L.foldl
			     (fn ((f,arg),str) => 
				 if arg  = 0
				 then str
				 else
				     (str ^ (Fun.toString f ^ 
					    (ListUtil.toStringCommaRound
						 (fn j => Int.toString (lookupVal j))
						 (L.tabulate (arg, fn i=> lookupFpsMap (fpsMap,(f,i+1)))))
					    )))
			     ""
			     fargs
			 else ""

	       val _ = if result then print (fps ^ "; ") else ()
	       val _ = if result then print ((#2 (countOrder lenRs)) ^ "\n") else ()


	   (* 	       val _ = if result  *)
	   (* 		       then (print "Decreasing by: "; printInfo resultArray encoding) *)
	   (* 		       else () *)
	   in 
	       result
	   end
       else false


   fun hasDecreasingCps smtSolverPath tmpDir rs =
       case (!decreasingDiagramVersion) of
	   1 => hasDecreasingCps0 smtSolverPath tmpDir rs
	 | 2 => hasDecreasingCps1 smtSolverPath tmpDir rs
	 | 3 => hasDecreasingCps1 smtSolverPath tmpDir rs
	 | 4 => hasDecreasingCps1 smtSolverPath tmpDir rs
	 | _ => hasDecreasingCps2 smtSolverPath tmpDir rs


   end (* local *)
end (* struct *)

(* for test 

val rs = IOFotrs.rdRules [ "nats -> cons(0,inc(nats))",  
			   "inc(cons(?x,?ys)) -> cons(s(?x),inc(?ys))",
			   "hd(cons(?x,?ys)) -> ?x",
			   "tl(cons(?x,?ys)) -> ?ys",
			   "inc(tl(nats)) -> tl(inc(nats))"
			 ];

val rs20 = IOFotrs.rdRules [ "g(a) -> f(g(a))",  
			     "g(b) -> c",
			     "a -> b",
			     "f(?x) -> h(?x,?x)",
			     "h(?x,?y) -> c"
			   ];


val rs21 = IOFotrs.rdRules [ "b -> a",
			     "b -> c",
			     "c -> h(b)",
			     "c -> d",
			     "a -> h(a)",
			     "d -> h(d)" ];


val minisatPath = "../../work/tools/bin/minisat"

val tmpDir = "../../work/rwchecker/"


*)
